import json
import logging
import sys
from datetime import datetime, timedelta

from django.utils.timezone import make_aware

from apps.nbworker.utils import WorkerState
from apps.notebooks.models import Notebook
from apps.ws.models import Worker
from apps.ws.utils import machine_uuid

log = logging.getLogger(__name__)


class DBClient:
    def __init__(self, notebook_id, session_id, worker_id):
        self.notebook_id = notebook_id
        self.session_id = session_id
        self.worker_id = worker_id
        self.worker = None  # db object
        self.state = WorkerState.Unknown
        self.notebook = None
        self.load_notebook()

    def load_notebook(self):
        try:
            log.debug(f"Load notebook id={self.notebook_id}")
            self.notebook = Notebook.objects.get(pk=self.notebook_id)
        except Exception:
            log.exception("Expetion when notebook load, quit")
            sys.exit(0)

    def is_presentation(self):
        try:
            log.debug(f"Check if notebook is presentation {self.notebook.output}")
            return self.notebook.output == "slides"
        except Exception:
            log.exception("Exception when check if notebook is presentation")
        return False

    def show_code(self):
        try:
            show_it = str(
                json.loads(self.notebook.params).get("show-code", "false")
            ).lower()
            log.debug(f"Check if show code from notebook ({show_it})")
            return show_it == "true"
        except Exception:
            log.exception("Exception when check if show code from notebook")
        return False

    def show_prompt(self):
        try:
            show_it = str(
                json.loads(self.notebook.params).get("show-prompt", "false")
            ).lower()
            log.debug(f"Check if show prompt from notebook ({show_it})")
            return show_it == "true"
        except Exception:
            log.exception("Exception when check if show promtp from notebook")
        return False

    def reveal_theme(self):
        # TODO: get reveal theme
        return "simple"

    def worker_state(self):
        return self.state

    def set_worker_state(self, new_state):
        try:
            log.debug(
                f"Worker id={self.worker_id} set state {new_state} uuid {machine_uuid()}"
            )
            self.state = new_state
            if self.worker_exists() and self.worker is not None:
                self.worker.state = new_state
                # set worker machine id
                # to control number of workers
                # in the single machine
                self.worker.machine_id = machine_uuid()
                self.worker.save()
        except Exception:
            log.exception("Exception when set worker state")

    @staticmethod
    def delete_worker_in_db(worker_id):
        try:
            log.debug(f"Delete worker id={worker_id}")
            Worker.objects.get(pk=worker_id).delete()
        except Exception:
            pass
            #log.exception(f"Exception when delete worker")

    def delete_worker(self):
        DBClient.delete_worker_in_db(self.worker_id)

    def delete_stale_workers(self):
        try:
            log.debug(
                (
                    "Delete stale workers, "
                    f"notebook_id={self.notebook_id}, "
                    f"session_id={self.session_id}, "
                    f"worker_id<{self.worker_id}"
                )
            )
            workers = Worker.objects.filter(
                session_id=self.session_id,
                notebook__id=self.notebook_id,
                pk__lt=self.worker_id,
            )
            workers.delete()
        except Exception as e:
            log.exception("Exception when delete stale workers")

    def worker_exists(self):
        try:
            log.debug(f"Worker id={self.worker_id} exists")
            self.worker = Worker.objects.get(pk=self.worker_id)
        except Worker.DoesNotExist as e:
            #log.exception(f"Worker id={self.worker_id} does not exists, quit")
            sys.exit(1)
        return True

    def is_worker_stale(self):
        try:
            log.debug(f"Check worker id={self.worker_id} is stale")
            self.worker = Worker.objects.get(pk=self.worker_id)
            return self.worker.updated_at < make_aware(
                datetime.now() - timedelta(minutes=1)
            )
        except Exception:
            log.exception(
                f"Exception when check if worker id={self.worker_id} is stale"
            )
        return True



import os
import tempfile

import nbformat as nbf
from django.contrib.auth.models import User
from django.test import Client, TestCase
from pro.accounts.models import Membership, MercuryGroup
from rest_framework.reverse import reverse

from apps.notebooks.models import Notebook
from apps.notebooks.tasks import task_init_notebook

# python manage.py test apps


def create_notebook_without_yaml(filename, text="# Title", code="print(1)"):
    nb = nbf.v4.new_notebook()

    nb["cells"] = [nbf.v4.new_markdown_cell(text), nbf.v4.new_code_cell(code)]
    with open(filename, "w", encoding="utf-8", errors="ignore") as f:
        nbf.write(nb, f)


def create_notebook_with_yaml(
    filename, yaml="", text="# Title in md", code="print(1)", code2=""
):
    nb = nbf.v4.new_notebook()
    text = "# Title"
    nb["cells"] = [
        nbf.v4.new_raw_cell(yaml),
        nbf.v4.new_markdown_cell(text),
        nbf.v4.new_code_cell(code),
    ]
    if code2 != "":
        nb["cells"] += [nbf.v4.new_code_cell(code2)]

    with open(filename, "w", encoding="utf-8", errors="ignore") as f:
        nbf.write(nb, f)


class InitNotebookTestCase(TestCase):
    def test_init_notebook_without_yaml(self):

        with tempfile.NamedTemporaryFile() as tmp:
            create_notebook_without_yaml(tmp.name + ".ipynb")
            task_init_notebook(tmp.name + ".ipynb")
            # in the case of missing YAML
            # notebook title should be 'Please provide title'
            nb = Notebook.objects.get(pk=1)
            self.assertEqual(nb.title, "Please provide title")

    def test_task_init_notebook(self):
        task_init_notebook(
            os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                "fixtures",
                "simple_notebook.ipynb",
            )
        )

    def test_init_notebook_with_empty_title(self):

        with tempfile.NamedTemporaryFile() as tmp:
            yaml = """---
title:
---"""
            create_notebook_with_yaml(tmp.name + ".ipynb", yaml)
            task_init_notebook(tmp.name + ".ipynb")

            # in the case of missing YAML
            # notebook title should be 'Please provide title'
            nb = Notebook.objects.get(pk=1)
            self.assertEqual(nb.title, "Please provide title")


class ShareNotebookTestCase(TestCase):
    def test_share_public(self):

        yaml = """---
share: public
---"""
        with tempfile.NamedTemporaryFile() as tmp:
            create_notebook_without_yaml(tmp.name + ".ipynb")
            task_init_notebook(tmp.name + ".ipynb")

        response = self.client.get("/api/v1/notebooks/")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 1)

    def test_share_private(self):

        self.assertEqual(Notebook.objects.all().count(), 0)

        yaml = """---
share: private
---"""
        with tempfile.NamedTemporaryFile() as tmp:
            create_notebook_with_yaml(tmp.name + ".ipynb", yaml=yaml)
            task_init_notebook(tmp.name + ".ipynb")

        # not logged user should see 0 notebooks
        response = self.client.get("/api/v1/notebooks/")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 0)

        # create user and login
        user = {
            "username": "piotrek",
            "password": "verysecret",
        }
        User.objects.create_user(username=user["username"], password=user["password"])
        response = self.client.post(
            reverse("rest_login"), user, content_type="application/json"
        )
        token = response.json()["key"]
        headers = {"HTTP_AUTHORIZATION": "Token " + token}
        # logged user with token in headers should see 1 notebook
        response = self.client.get("/api/v1/notebooks/", **headers)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 1)

    def test_share_username(self):

        # create user and login
        user = {
            "username": "piotrek",
            "password": "verysecret",
        }
        User.objects.create_user(username=user["username"], password=user["password"])
        response = self.client.post(
            reverse("rest_login"), user, content_type="application/json"
        )
        token = response.json()["key"]
        headers = {"HTTP_AUTHORIZATION": "Token " + token}

        self.assertEqual(Notebook.objects.all().count(), 0)

        yaml = f"""---
share: {user["username"]}
---"""
        with tempfile.NamedTemporaryFile() as tmp:
            create_notebook_with_yaml(tmp.name + ".ipynb", yaml=yaml)
            task_init_notebook(tmp.name + ".ipynb")

        # not logged user should see 0 notebooks
        response = self.client.get("/api/v1/notebooks/")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 0)

        # logged should see 1 notebook
        response = self.client.get("/api/v1/notebooks/", **headers)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 1)

        # create second user
        user2 = {
            "username": "piotrek2",
            "password": "verysecret2",
        }
        User.objects.create_user(username=user2["username"], password=user2["password"])
        response = self.client.post(
            reverse("rest_login"), user2, content_type="application/json"
        )
        token2 = response.json()["key"]
        headers2 = {"HTTP_AUTHORIZATION": "Token " + token2}

        # logged by different user should see 0 notebooks
        response = self.client.get("/api/v1/notebooks/", **headers2)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 0)

    def test_share_group(self):

        # create user and login
        user = {
            "username": "piotrek",
            "password": "verysecret",
        }
        user_obj1 = User.objects.create_user(
            username=user["username"], password=user["password"]
        )
        response = self.client.post(
            reverse("rest_login"), user, content_type="application/json"
        )
        token = response.json()["key"]
        headers = {"HTTP_AUTHORIZATION": "Token " + token}

        # create second user
        user2 = {
            "username": "piotrek2",
            "password": "verysecret2",
        }
        user_obj2 = User.objects.create_user(
            username=user2["username"], password=user2["password"]
        )
        response = self.client.post(
            reverse("rest_login"), user2, content_type="application/json"
        )
        token2 = response.json()["key"]
        headers2 = {"HTTP_AUTHORIZATION": "Token " + token2}

        # create third user
        user3 = {
            "username": "piotrek3",
            "password": "verysecret3",
        }
        user_obj3 = User.objects.create_user(
            username=user3["username"], password=user3["password"]
        )
        response = self.client.post(
            reverse("rest_login"), user3, content_type="application/json"
        )
        token3 = response.json()["key"]
        headers3 = {"HTTP_AUTHORIZATION": "Token " + token3}

        #
        # create a group and add first and second used
        #
        group = MercuryGroup(name="group1")
        group.save()

        Membership(user=user_obj1, group=group).save()
        Membership(user=user_obj2, group=group).save()

        self.assertEqual(Notebook.objects.all().count(), 0)

        yaml = f"""---
share: {group.name}
---"""
        with tempfile.NamedTemporaryFile() as tmp:
            create_notebook_with_yaml(tmp.name + ".ipynb", yaml=yaml)
            task_init_notebook(tmp.name + ".ipynb")

        # not logged user should see 0 notebooks
        response = self.client.get("/api/v1/notebooks/")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 0)

        # logged as first user should see 1 notebook
        response = self.client.get("/api/v1/notebooks/", **headers)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 1)

        # logged as second user should see 1 notebook
        response = self.client.get("/api/v1/notebooks/", **headers2)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 1)

        # logged as third user should see 0 notebooks
        response = self.client.get("/api/v1/notebooks/", **headers3)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 0)

    def test_get_single_notebook(self):
        # create user and login
        user = {
            "username": "piotrek",
            "password": "verysecret",
        }
        User.objects.create_user(username=user["username"], password=user["password"])
        response = self.client.post(
            reverse("rest_login"), user, content_type="application/json"
        )
        token = response.json()["key"]
        headers = {"HTTP_AUTHORIZATION": "Token " + token}

        # create second user
        user2 = {
            "username": "piotrek2",
            "password": "verysecret2",
        }
        user_obj2 = User.objects.create_user(
            username=user2["username"], password=user2["password"]
        )
        response = self.client.post(
            reverse("rest_login"), user2, content_type="application/json"
        )
        token2 = response.json()["key"]
        headers2 = {"HTTP_AUTHORIZATION": "Token " + token2}

        self.assertEqual(Notebook.objects.all().count(), 0)

        yaml = f"""---
share: {user2["username"]}
---"""
        with tempfile.NamedTemporaryFile() as tmp:
            create_notebook_with_yaml(tmp.name + ".ipynb", yaml=yaml)
            task_init_notebook(tmp.name + ".ipynb")

        yaml = f"""---
share: private
---"""
        with tempfile.NamedTemporaryFile() as tmp:
            create_notebook_with_yaml(tmp.name + ".ipynb", yaml=yaml)
            task_init_notebook(tmp.name + ".ipynb")

        # first user should see see 1 notebook
        response = self.client.get("/api/v1/notebooks/", **headers)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 1)

        # second user should see see 2 notebooks
        response = self.client.get("/api/v1/notebooks/", **headers2)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 2)

        # first user can not get first notebook
        response = self.client.get("/api/v1/notebooks/1/", **headers)
        self.assertEqual(response.status_code, 404)
        # first user can get first notebook
        response = self.client.get("/api/v1/notebooks/2/", **headers)
        self.assertEqual(response.status_code, 200)

        # second user can access both notebooks
        for i in [1, 2]:
            response = self.client.get(f"/api/v1/notebooks/{i}/", **headers2)
            self.assertEqual(response.status_code, 200)



import os

from django.conf import settings
from rest_framework.response import Response
from rest_framework.views import APIView

from server.settings import is_pro


class VersionInfo(APIView):
    def get(self, request, format=None):
        return Response({"isPro": is_pro})


# it will be used as really simple cache
welcome_msg = None


class WelcomeMessage(APIView):
    def get(self, request, format=None):
        global welcome_msg
        welcome_file = os.environ.get("WELCOME")
        if welcome_file is not None:
            if welcome_msg is None:
                if os.path.exists(welcome_file):
                    with open(welcome_file, encoding="utf-8", errors="ignore") as fin:
                        welcome_msg = fin.read()
                else:
                    return Response({"msg": ""})
            if welcome_msg is not None:
                return Response({"msg": welcome_msg})
        return Response({"msg": ""})



from django.db import models

from apps.notebooks.models import Notebook


class Task(models.Model):
    # task id from Celery
    task_id = models.CharField(max_length=128, blank=True)
    # web browser session id
    session_id = models.CharField(max_length=128)
    # notebook
    notebook = models.ForeignKey(
        Notebook,
        on_delete=models.CASCADE,
    )
    created_at = models.DateTimeField(auto_now_add=True)
    # state of execution, can be: CREATED, RECEIVED, DONE, ERROR
    state = models.CharField(max_length=128, blank=True)
    # input params for task
    params = models.TextField(blank=True)
    # result of execution, should contain
    # the path with HTML notebook
    result = models.TextField(blank=True)



import json
import os
import queue
import signal
import sys
import threading
import time
import traceback
from datetime import datetime, timedelta

import websocket
from django.utils.timezone import make_aware

CURRENT_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
BACKEND_DIR = os.path.join(CURRENT_DIR, "..")
sys.path.insert(0, BACKEND_DIR)

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "server.settings")
import django

django.setup()


import logging

LOG_LEVEL = logging.ERROR #DEBUG
logging.basicConfig(
    # filename="nbworker.log", filemode="w",
    format="NB %(asctime)s %(message)s",
    level=LOG_LEVEL,
)
log = logging.getLogger(__name__)

from apps.nbworker.utils import stop_event
from apps.nbworker.db import DBClient
from apps.nbworker.nb import NBWorker

if len(sys.argv) != 5:
    log.error("Wrong number of input parameters")
    sys.exit(0)

notebook_id = int(sys.argv[1])
session_id = sys.argv[2]
worker_id = int(sys.argv[3])
server_url = sys.argv[4]

def signal_handler(signal, frame):
    global stop_event
    log.debug("\nBye bye!")
    stop_event.set()
    DBClient.delete_worker_in_db(worker_id)
    sys.exit(1)


signal.signal(signal.SIGINT, signal_handler)


RECONNECT_WAIT_TIME = 10
CONNECT_MAX_TRIES = 2


if __name__ == "__main__":
    log.info(f"Start NBWorker with arguments {sys.argv}")
    for _ in range(CONNECT_MAX_TRIES):
        nb_worker = NBWorker(
            f"{server_url}/ws/worker/{notebook_id}/{session_id}/{worker_id}/",
            notebook_id,
            session_id,
            worker_id,
        )
        time.sleep(RECONNECT_WAIT_TIME)



# Generated by Django 3.2.5 on 2022-06-01 09:10

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Notebook",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("file_updated_at", models.DateTimeField(blank=True)),
                ("title", models.CharField(max_length=512)),
                ("slug", models.CharField(blank=True, max_length=512)),
                ("path", models.CharField(max_length=1024)),
                ("share", models.TextField(blank=True)),
                ("params", models.TextField(blank=True)),
                ("state", models.CharField(blank=True, max_length=128)),
                ("default_view_path", models.CharField(blank=True, max_length=1024)),
                ("output", models.CharField(blank=True, max_length=128)),
                ("format", models.CharField(blank=True, max_length=1024)),
                ("schedule", models.CharField(blank=True, max_length=128)),
                ("notify", models.TextField(blank=True)),
                ("errors", models.TextField(blank=True)),
            ],
        ),
    ]



import json
import os
import tempfile

import ipywidgets
from IPython.display import display

from .manager import WidgetsManager


class DirPath:
    def __init__(self, dir_path):
        self.value = dir_path


class OutputDir:
    def __init__(self):
        self.code_uid = WidgetsManager.get_code_uid("OutputDir")
        if WidgetsManager.widget_exists(self.code_uid):
            self.dir_path = WidgetsManager.get_widget(self.code_uid)
        else:
            self.dir_path = DirPath(".")
            WidgetsManager.add_widget("output-dir", self.code_uid, self.dir_path)
        display(self)

    @property
    def path(self):
        return self.dir_path.value

    def __str__(self):
        return "mercury.OutputDir"

    def __repr__(self):
        return "mercury.OutputDir"

    def _repr_mimebundle_(self, **kwargs):

        data = {}

        view = {
            "widget": "OutputDir",
            "model_id": "output-dir",
            "code_uid": self.code_uid,
        }
        data["application/mercury+json"] = json.dumps(view, indent=4)
        data["text/html"] = "<h3>Output Directory</h3><small>This output won't appear in the web app.</small>"
        

        return data



import os

from celery import shared_task
from django.conf import settings

from apps.notebooks.models import Notebook
from apps.tasks.export_pdf import to_pdf


@shared_task(bind=True)
def export_to_pdf(self, job_params):
    notebook_id = job_params.get("notebook_id")
    notebook_path = job_params.get("notebook_path")

    if notebook_id is None or notebook_path is None:
        raise Exception(
            "PDF export params validation error. Wrong notebook information."
        )

    # try to build platform independent path

    notebook_os_path = os.path.join(
        *(
            [settings.MEDIA_ROOT]
            + notebook_path.replace(settings.MEDIA_URL, "", 1).split("/")
        )
    )

    if not os.path.exists(notebook_os_path):
        raise Exception(
            f"PDF export notebook error. The notebook in HTML format does not exist."
        )

    notebook = Notebook.objects.get(pk=notebook_id)

    slides_postfix = ""
    if notebook.output == "slides":
        slides_postfix = "?print-pdf"

    pdf_os_path = notebook_os_path.replace(".html", ".pdf")

    to_pdf(notebook_os_path + slides_postfix, pdf_os_path)

    title = notebook.slug + ".pdf"

    pdf_url = notebook_path.replace(".html", ".pdf")

    return pdf_url, title



import logging
import os
import subprocess
import sys

from celery import shared_task
from django.conf import settings

from apps.ws.models import Worker
from apps.ws.utils import machine_uuid

log = logging.getLogger(__name__)


@shared_task(bind=True)
def task_start_websocket_worker(self, job_params):

    log.debug(f"NbWorkers per machine: {settings.NBWORKERS_PER_MACHINE}")
    machine_id = machine_uuid()
    workers = Worker.objects.filter(machine_id=machine_id)

    log.debug(f"Workers count: {len(workers)} machine_id={ machine_id }")

    if len(workers) > settings.NBWORKERS_PER_MACHINE:
        log.debug("Defer task start ws worker")
        task_start_websocket_worker.s(job_params).apply_async(countdown=15)
    else:
        directory = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        command = [
            sys.executable,
            os.path.join(directory, "nbworker"),
            str(job_params["notebook_id"]),
            str(job_params["session_id"]),
            str(job_params["worker_id"]),
            job_params["server_url"],
        ]
        log.debug("Start " + " ".join(command))
        worker = subprocess.Popen(command)



from rest_framework import serializers

from apps.notebooks.models import Notebook


class NotebookSerializer(serializers.ModelSerializer):
    class Meta:
        model = Notebook
        read_only_fields = ("id", "created_at", "file_updated_at")
        fields = (
            "id",
            "created_at",
            "file_updated_at",
            "title",
            "slug",
            "path",
            "share",
            "params",
            "state",
            "default_view_path",
            "output",
            "format",
            "schedule",
            "errors",
        )



import json
import logging
import os
import subprocess
import sys
import uuid

from datetime import datetime
from shutil import which
from subprocess import PIPE, Popen

import nbformat
import yaml
from celery import shared_task
from croniter import croniter
from django.conf import settings
from django.template.defaultfilters import slugify
from django.utils.timezone import make_aware

from apps.nb.exporter import Exporter
from apps.notebooks.models import Notebook
from apps.notebooks.slides_themes import SlidesThemes
from apps.tasks.models import Task
from apps.tasks.notify import validate_notify
from apps.ws.utils import parse_params

log = logging.getLogger(__name__)


def process_nbconvert_errors(error_msg):
    known_warnings = [
        "warn(",
        "UserWarning",
        "FutureWarning",
        "[NbConvertApp] Converting notebook",
        "[NbConvertApp] Writing",
    ]
    error_lines = []
    for e in error_msg.decode("utf-8").split("\n"):
        if e == "":
            continue
        known_warning = False
        for w in known_warnings:
            if w in e:
                known_warning = True
                break
        if not known_warning and e != "":
            error_lines += [e]
    return "\n".join(error_lines)


def get_hash():
    h = uuid.uuid4().hex.replace("-", "")
    return h[:5]


def is_tool(name):
    return which(name) is not None


def get_jupyter_bin_path():
    if is_tool("jupyter"):
        return "jupyter"
    if is_tool("jupyter.exe"):
        return "jupyter"
    if sys.executable.endswith("exe"):
        return os.path.join(os.path.dirname(sys.executable), "Scripts", "jupyter.exe")
    return os.path.join(os.path.dirname(sys.executable), "jupyter")


def available_kernels():
    command = [
        get_jupyter_bin_path(),
        "kernelspec",
        "list",
    ]
    msg = []
    with Popen(command, stdout=PIPE) as proc:
        msg = proc.stdout.read().decode("utf-8").split("\n")
    if not msg:
        return []
    kernels = []
    start_parsing = False
    for m in msg:
        if start_parsing:
            parts = m.split(" ")
            parts = [p for p in parts if p != ""]
            if len(parts) == 2:
                kernels += [parts[0]]  # parts[1] is a kernel path
        if "Available kernels:" in m:
            start_parsing = True
    # list of kernels name
    return kernels


def nb_default_title(nb_path):
    try:
        fname = os.path.basename(nb_path)
        if "." in fname:
            return ".".join(fname.split(".")[:-1])
        return fname
    except Exception as e:
        log.exception("Problem when get default title from notebook")

    return "Please provide title"


def is_presentation(nb):
    for cell in nb.cells:
        if "slideshow" in cell.get("metadata", {}):
            return True
    return False


def task_init_notebook(
    notebook_path, render_html=True, is_watch_mode=False, notebook_id=None
):

    try:

        kernels = available_kernels()

        params = {
            "title": "",
            "author": "Please provide author",
            "description": "Please provide description",
            "share": "public",
            "output": "app",
            "format": {},
            "schedule": "",
            "notify": {},
        }
        nb = None
        update_notebook = False
        with open(notebook_path, encoding="utf-8", errors="ignore") as f:
            nb = nbformat.read(f, as_version=4)
            if "cells" in nb and len(nb["cells"]) > 0:
                first_cell = nb["cells"][0]["source"]
                if first_cell.startswith("---") and first_cell.endswith("---"):
                    params = yaml.safe_load(first_cell[3:-3])

            if (
                "metadata" in nb
                and "kernelspec" in nb["metadata"]
                and "name" in nb["metadata"]["kernelspec"]
            ):
                kernel_name = nb["metadata"]["kernelspec"]["name"]
                if kernel_name not in kernels:
                    print("*" * 42)
                    print(f"Your notebook kernel name is set to '{kernel_name}'.")
                    print(f"In this system available kernels are {kernels}.")

                    if len(kernels) > 0:
                        print(f"The script will change the kernel name to {kernels[0]}")
                        new_kernel_name = kernels[0]
                        nb["metadata"]["kernelspec"]["name"] = new_kernel_name
                        update_notebook = True
                    else:
                        print(
                            "Sorry, cant automatically update the kernel name in the notebook."
                        )
                        return

            # check if nb in V2
            parse_params(nb, params)

        if nb is None:
            raise Exception(f"Cant read notebook from {notebook_path}")

        if update_notebook and nb is not None:
            with open(notebook_path, "w", encoding="utf-8", errors="ignore") as f:
                nbformat.write(nb, f)

        if "date" in params:
            params["date"] = str(params["date"])

        notebook_title = params.get("title", "")
        if notebook_title == "":
            notebook_title = nb_default_title(notebook_path)
        notebook_share = params.get("share", "public")
        notebook_output = params.get("output", "app")
        notebook_format = params.get("format", {})
        notebook_schedule = params.get("schedule", "")
        notebook_notify = params.get("notify", {})

        if is_presentation(nb):
            # automatically detect slides in cells
            # and set slides output
            notebook_output = "slides"

        if notebook_schedule != "":
            try:
                croniter.is_valid(notebook_schedule)
            except Exception as e:
                raise Exception(
                    f"The schedule ({notebook_schedule}) is not valid. {str(e)} Please check schedule at https://crontab.guru"
                )

        # make sure that there are commas and no spaces between commas
        notebook_share = (
            "," + ",".join([i.strip() for i in notebook_share.split(",")]) + ","
        )

        notebook_slug = params.get("slug", "")
        if notebook_slug == "":
            notebook_slug = slugify(notebook_title)
        notebook_output_file = notebook_slug
        if notebook_id is not None:
            notebook_output_file = f"{notebook_slug}-{get_hash()}"

        if render_html:

            exporter = Exporter(
                show_code=params.get("show-code", False),
                show_prompt=params.get("show-prompt", False),
                is_presentation=notebook_output == "slides",
                reveal_theme=notebook_format.get("theme", "white"),
            )
            body = exporter.export(nb)

            with open(
                os.path.join(settings.MEDIA_ROOT, f"{notebook_output_file}.html"), "w"
            ) as fout:
                fout.write(body)

            """
            command = [
                get_jupyter_bin_path(),
                "nbconvert",
                "--RegexRemovePreprocessor.patterns=^---",
                notebook_path,
                "--output",
                notebook_output_file,
                "--output-dir",
                settings.MEDIA_ROOT,
                "--to",
                "slides" if notebook_output == "slides" else "html",
                "--Application.log_level=40",  # 30 is default, 40 is ERROR, 50 is Critical
            ]
            if "show-code" in params and not params["show-code"]:
                command += ["--no-input"]
            if "show-prompt" in params and not params["show-prompt"]:
                command += ["--no-prompt"]

            if notebook_output == "slides":
                command += SlidesThemes.nbconvert_options(notebook_format)

            error_msg = ""
            with Popen(command, stdout=PIPE, stderr=PIPE) as proc:
                # print(proc.stdout.read())
                # print(proc.stderr.read())
                error_msg = proc.stderr.read()

            error_msg = process_nbconvert_errors(error_msg)
            if error_msg != "":
                print(error_msg)
            """
            error_msg = ""  # TODO: handle errors

            """
            # change file name if needed
            if notebook_output == "slides":
                expected_fpath = os.path.join(
                    settings.MEDIA_ROOT, f"{notebook_output_file}.html"
                )
                slides_fpath = os.path.join(
                    settings.MEDIA_ROOT, f"{notebook_output_file}.slides.html"
                )
                if os.path.exists(slides_fpath):
                    os.rename(slides_fpath, expected_fpath)
            """
            if not params.get("show-code", False):  # "--no-input" in command:
                with open(
                    os.path.join(settings.MEDIA_ROOT, f"{notebook_output_file}.html"),
                    "a",
                    encoding="utf-8",
                    errors="ignore",
                ) as fout:
                    fout.write(
                        """\n<style type="text/css">
.jp-mod-noOutputs {
    padding: 0px; 
}
.jp-mod-noInput {
  padding-top: 0px;
  padding-bottom: 0px;
}
</style>"""
                    )

            """
            if notebook_output == "slides":
                with open(
                    os.path.join(settings.MEDIA_ROOT, f"{notebook_output_file}.html"),
                    "a",
                    encoding="utf-8",
                    errors="ignore",
                ) as fout:
                    fout.write(SlidesThemes.additional_css(notebook_format))
            """

        parse_errors = validate_notify(notebook_notify)

        if notebook_id is None:
            notebook = Notebook(
                title=notebook_title,
                slug=notebook_slug,
                path=os.path.abspath(notebook_path),
                state="WATCH_READY" if is_watch_mode else "READY",
                share=notebook_share,
                params=json.dumps(params),
                default_view_path=os.path.join(
                    settings.MEDIA_URL, f"{notebook_output_file}.html"
                ),
                file_updated_at=make_aware(
                    datetime.fromtimestamp(os.path.getmtime(notebook_path))
                ),
                output=notebook_output,
                format=json.dumps(notebook_format),
                schedule=notebook_schedule,
                notify=json.dumps(notebook_notify),
                errors=parse_errors,
            )
        else:

            notebook = Notebook.objects.get(pk=notebook_id)
            notebook.title = notebook_title
            notebook.slug = notebook_slug
            notebook.path = os.path.abspath(notebook_path)
            notebook.state = "WATCH_READY" if is_watch_mode else "READY"
            notebook.share = notebook_share
            notebook.params = json.dumps(params)
            # remove old default view
            if os.path.exists(notebook.default_view_path):
                os.remove(notebook.default_view_path)
            notebook.default_view_path = os.path.join(
                settings.MEDIA_URL, f"{notebook_output_file}.html"
            )
            notebook.file_updated_at = make_aware(
                datetime.fromtimestamp(os.path.getmtime(notebook_path))
            )
            notebook.output = notebook_output
            notebook.format = json.dumps(notebook_format)
            notebook.schedule = notebook_schedule
            notebook.notify = json.dumps(notebook_notify)
            notebook.errors = parse_errors

        notebook.save()
        return notebook.id
    except KeyboardInterrupt:
        try:
            sys.exit(0)
        except SystemExit:
            os._exit(0)
    except Exception as e:
        if notebook_id is not None:
            raise e
        else:
            print("Error during notebook initialization.", str(e))


@shared_task(bind=True)
def task_watch(self, notebook_id):

    notebook = None
    try:
        notebook = Notebook.objects.get(pk=notebook_id)
        current_update_time = datetime.fromtimestamp(os.path.getmtime(notebook.path))
        notebook_updated_at = notebook.file_updated_at.replace(tzinfo=None)

        if current_update_time != notebook_updated_at and notebook.state in [
            "WATCH_READY",
            "WATCH_ERROR",
        ]:
            notebook.state = "WATCH_WAIT"
            notebook.save()
            # clear all tasks
            Task.objects.filter(notebook__id=notebook.id).delete()
            # initialize updated notebook
            task_init_notebook(
                notebook.path,
                render_html=True,
                is_watch_mode=True,
                notebook_id=notebook.id,
            )
    except Exception as e:
        if notebook is not None:
            notebook.state = "WATCH_ERROR"
            notebook.save()



import os
import tempfile
import nbformat as nbf

def create_demo_notebook(filename = "demo.ipynb"):
    nb = nbf.v4.new_notebook()
    
    imports = "import mercury as mr"
    text = "# 🚀 Demo notebook with greetings 👋"
    variables = '''year = mr.Slider(label="Please select the year", value=2023, min=2000, max=2050)
name = mr.Text(label="What is your name?", value="Piotr")'''
    code = '''print(f"Hello {name.value} in {year.value}!")'''
    nb["cells"] = [ 
        nbf.v4.new_code_cell(imports), 
        nbf.v4.new_markdown_cell(text), 
        nbf.v4.new_code_cell(variables), 
        nbf.v4.new_code_cell(code)
    ]
    with open(filename, "w", encoding="utf-8", errors="ignore") as f:
        nbf.write(nb, f)





from django.conf.urls import include, url
from rest_framework.routers import DefaultRouter

from apps.tasks.views import (
    ClearTasksView,
    CreateRestAPITask,
    ExecutionHistoryView,
    ExportPDF,
    GetLastTaskView,
    GetPDFAddress,
    GetRestAPITask,
    ListOutputFilesView,
    ListWorkerOutputFilesView,
    TaskCreateView,
)

tasks_urlpatterns = [
    url("api/v1/execute/(?P<notebook_id>.+)", TaskCreateView.as_view()),
    url(
        "api/v1/latest_task/(?P<notebook_id>.+)/(?P<session_id>.+)",
        GetLastTaskView.as_view(),
    ),
    url(
        "api/v1/output_files/(?P<session_id>.+)/(?P<task_id>.+)",
        ListOutputFilesView.as_view(),
    ),
    url(
        "api/v1/worker_output_files/(?P<session_id>.+)/(?P<worker_id>.+)",
        ListWorkerOutputFilesView.as_view(),
    ),
    url(
        "api/v1/clear_tasks/(?P<notebook_id>.+)/(?P<session_id>.+)",
        ClearTasksView.as_view(),
    ),
    # used by notebook as REST API
    url("run/(?P<notebook_slug>.+)", CreateRestAPITask.as_view()),
    url("get/(?P<session_id>.+)", GetRestAPITask.as_view()),
    url("export_pdf", ExportPDF.as_view()),
    url("get_pdf/(?P<job_id>.+)", GetPDFAddress.as_view()),
    url(
        "api/v1/execution_history/(?P<notebook_id>.+)/(?P<session_id>.+)",
        ExecutionHistoryView.as_view(),
    ),
]



import nbformat as nbf


def test_notebook(markdown=[], code=[]):
    nb = nbf.v4.new_notebook()
    nb["cells"] = []
    for m in markdown:
        nb["cells"] += [nbf.v4.new_markdown_cell(m)]
    for c in code:
        nb["cells"] += [nbf.v4.new_code_cell(c)]

    return nb


def one_cell_notebook(code=""):
    nb = nbf.v4.new_notebook()
    nb["cells"] = [nbf.v4.new_code_cell(code)]
    return nb



#!/usr/bin/env python
"""Django's command-line utility for administrative tasks."""
import os
import sys
import django
import json
import subprocess
import webbrowser

from glob import glob
from django.core.management.utils import get_random_secret_key

CURRENT_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
BACKEND_DIR = os.path.join(CURRENT_DIR, "mercury")
sys.path.insert(0, BACKEND_DIR)

from demo import create_demo_notebook

__version__ = "1.99.1"

from widgets.manager import WidgetsManager
from widgets.app import App
from widgets.slider import Slider
from widgets.select import Select
from widgets.range import Range
from widgets.text import Text
from widgets.file import File
from widgets.checkbox import Checkbox
from widgets.numeric import Numeric
from widgets.multiselect import MultiSelect
from widgets.outputdir import OutputDir
from widgets.note import Note
from widgets.button import Button

VERBOSE = 0  # can be 0,1,2,3; 0 is no output


def main():
    """Run administrative tasks."""
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "server.settings")
    try:
        from django.core.management import execute_from_command_line
    except ImportError as exc:
        raise ImportError(
            "Couldn't import Django. Are you sure it's installed and "
            "available on your PYTHONPATH environment variable? Did you "
            "forget to activate a virtual environment?"
        ) from exc

    run_add_notebook = None
    if "run" in sys.argv:
        if os.environ.get("ALLOWED_HOSTS") is None:
            os.environ["ALLOWED_HOSTS"] = "*"
        if os.environ.get("SERVE_STATIC") is None:
            os.environ["SERVE_STATIC"] = "True"
        if os.environ.get("NOTEBOOKS") is None:
            os.environ["NOTEBOOKS"] = "*.ipynb"
        if os.environ.get("WELCOME") is None:
            os.environ["WELCOME"] = "welcome.md"
        if os.environ.get("SECRET_KEY") is None:
            os.environ["SECRET_KEY"] = get_random_secret_key()
        i = sys.argv.index("run")
        sys.argv[i] = "runserver"
        sys.argv.append("--runworker")
        logo = """                            

     _ __ ___   ___ _ __ ___ _   _ _ __ _   _ 
    | '_ ` _ \ / _ \ '__/ __| | | | '__| | | |
    | | | | | |  __/ | | (__| |_| | |  | |_| |
    |_| |_| |_|\___|_|  \___|\__,_|_|   \__, |
                                         __/ |
                                        |___/ 
        """
        print(logo)

        if "demo" in sys.argv:
            create_demo_notebook("demo.ipynb")
            sys.argv.remove("demo")
            run_add_notebook = "demo.ipynb"
        else:
            for l in sys.argv:
                if l.endswith(".ipynb"):
                    run_add_notebook = l

    if "--noadditional" not in sys.argv:
        execute_from_command_line(["mercury.py", "migrate", "-v", 0])

        superuser_username = os.environ.get("DJANGO_SUPERUSER_USERNAME")
        if (
            superuser_username is not None
            and os.environ.get("DJANGO_SUPERUSER_EMAIL") is not None
            and os.environ.get("DJANGO_SUPERUSER_PASSWORD") is not None
        ):
            try:
                from django.contrib.auth import get_user_model

                User = get_user_model()
                if not User.objects.filter(username=superuser_username):
                    execute_from_command_line(
                        ["mercury.py", "createsuperuser", "--noinput"]
                    )
            except Exception as e:
                print(str(e))
        if os.environ.get("SERVE_STATIC") is not None:
            execute_from_command_line(
                ["mercury.py", "collectstatic", "--noinput", "-v", "0"]
            )
        if os.environ.get("NOTEBOOKS") is not None and run_add_notebook is None:
            notebooks_str = os.environ.get("NOTEBOOKS")
            notebooks = []
            if "[" in notebooks_str or "{" in notebooks_str:
                notebooks = json.loads(notebooks_str)
            elif "*" in notebooks_str:
                notebooks = glob(notebooks_str)
            else:
                notebooks = [notebooks_str]
            for nb in notebooks:
                execute_from_command_line(["mercury.py", "add", nb])

        if run_add_notebook is not None:
            execute_from_command_line(["mercury.py", "add", run_add_notebook])
            if run_add_notebook in sys.argv:
                sys.argv.remove(run_add_notebook)

        worker = None
        if (
            os.environ.get("RUN_WORKER", "False") == "True"
            or "runworker" in sys.argv
            or "--runworker" in sys.argv
        ):
            worker_command = [
                "celery",
                "-A",
                "mercury.server" if sys.argv[0].endswith("mercury") else "server",
                "worker",
                "--loglevel=error",
                # "--loglevel=debug",
                "-P",
                "gevent",
                "--concurrency",
                "1",
                "-E",
            ]
            worker = subprocess.Popen(
                worker_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )

            # celery worker beat for periodic tasks
            beat_command = [
                "celery",
                "-A",
                "mercury.server" if sys.argv[0].endswith("mercury") else "server",
                "beat",
                "--loglevel=error",
                "--max-interval",
                "60",  # sleep 60 seconds
            ]
            subprocess.Popen(
                beat_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )

            if "--runworker" in sys.argv:
                sys.argv.remove("--runworker")

            if "runworker" in sys.argv:
                worker.wait()
    else:
        sys.argv.remove("--noadditional")

    try:

        arguments = sys.argv
        if (
            len(sys.argv) > 1
            and arguments[1] == "runserver"
            and "--noreload" not in arguments
        ):
            arguments += ["--noreload"]
            try:
                # open web browser if we are running a server
                url = "http://127.0.0.1:8000"
                webbrowser.open(url)
            except Exception as e:
                pass

        execute_from_command_line(arguments)

    except KeyboardInterrupt:
        try:
            sys.exit(0)
        except SystemExit:
            os._exit(0)
    except Exception as e:
        print("Mercury error.", str(e))


if __name__ == "__main__":
    main()
